from collections.abc import Sequence

from ampel.t1.T1SimpleCombiner import T1SimpleCombiner


class LSSTT1Combiner(T1SimpleCombiner):
    """Dummy class that exists only to provide default values for access and policy"""

    access: Sequence[int | str] = []
    policy: Sequence[int | str] = []



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/ingest/LSSTMongoMuxer.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.12.2017
# Last Modified Date: 18.03.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

import datetime
from typing import Any

from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.content.DataPoint import DataPoint
from ampel.types import StockId
from ampel.util.mappings import unflatten_dict


class LSSTMongoMuxer(AbsT0Muxer):
    """
    This class compares info between alert and DB so that only the needed info is ingested later.

    # Not used - removed
    :param alert_history_length: alerts must not contain all available info for a given transient.
    Alerts for LSST should provide a photometric history of 365 days.

    JN: For *elasticc* diffimage and forced photometry will have identical flux
    values. Since there is both a previous source phot and previous forced phot
    history, a single alert can contain multiple copes of the same measurement.
    Again, for *elasticc* it also seems like they keep the same id
    ForcedSourceID and SourceID
    for both kinds. As JN understand, this means that as new measurements come
    in, the typical SourceId flux values will be retained and forced phot
    "updates" not added to the DB. As the flux is the same this _should_ not
    matter.
    What we should do is probably:
    - Generate ID datapoints in LSSTDataPointShaper, prob of a hash from
    [time, flux, ccdVisitId(?)].
    - Make a proper superseeded search a la ZiMongoMuxer here where we would
    look for identical time (only?)
    But since this should not make a difference, provided we drop duplicates in
    the LSSTDataPointShaper it seems that this change is not strictly necessary
    now (assuming I got everything right.)


    """

    # Standard projection used when checking DB for existing PPS/ULS
    projection: dict[str, int] = {
        "_id": 0,
        "id": 1,
        "tag": 1,
        "channel": 1,
        "excl": 1,
        "stock": 1,
        "body.midpointMjdTai": 1,
        "body.band": 1,
        "body.psfFlux": 1,
        "body.diaObjectId": 1,
    }

    #: Require minimum time-to-live for datapoints
    min_ttl: None | float = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # used to check potentially already inserted pps
        self._photo_col = self.context.db.get_collection("t0")
        self._projection_spec = unflatten_dict(self.projection)

    def process(
        self, dps_al: list[DataPoint], stock_id: None | StockId = None
    ) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
        """
        :param dps_al: datapoints from alert
        :param stock_id: stock id from alert
        Attempt to determine which pps/uls should be inserted into the t0 collection,
        and which one should be marked as superseded.
        """

        # Part 1: gather info from DB and alert
        #######################################

        # New pps/uls lists for db loaded datapoints
        filter: dict[str, Any] = {"stock": stock_id}
        if self.min_ttl is not None:
            filter["expiry"] = {
                "$not": {
                    "$lt": datetime.datetime.now(tz=datetime.timezone.utc)
                    + datetime.timedelta(seconds=self.min_ttl)
                }
            }
        dps_db: list[DataPoint] = list(
            self._photo_col.find(filter, self.projection)
        )

        # Create set with datapoint ids from alert
        ids_dps_alert = {el["id"]: el for el in dps_al}

        # python set of ids of datapoints from DB
        ids_dps_db = {el["id"]: el for el in dps_db}

        # Part 2: Insert new data points
        ################################

        # Difference between candids from the alert and candids present in DB
        ids_dps_to_insert = ids_dps_alert.keys() - ids_dps_db.keys()

        # Emit datapoints in the order given in the alert, but prefer content
        # from the database. This allows the ingestion handler to detect when an
        # additional channel accepts a datapoint that was already in the
        # database.
        dps_combine = [
            ids_dps_db[dp["id"]]
            if dp["id"] in ids_dps_db
            else self._project(ids_dps_alert[dp["id"]], self.projection)
            for dp in dps_al
        ]

        return [
            dp for dp in dps_al if dp["id"] in ids_dps_to_insert
        ], dps_combine

    def _project(self, doc, projection):
        out: dict[str, Any] = {}
        for key, spec in projection.items():
            if key not in doc:
                continue

            if isinstance(spec, dict):
                item = doc[key]
                if isinstance(item, list):
                    out[key] = [self._project(v, spec) for v in item]
                elif isinstance(item, dict):
                    out[key] = self._project(item, spec)
            else:
                out[key] = doc[key]

        return out



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/alert/LSSTAlertSupplier.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.04.2021
# Last Modified Date: 21.03.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Iterator
from itertools import chain
from typing import Literal

from ampel.alert.AmpelAlert import AmpelAlert
from ampel.alert.BaseAlertSupplier import BaseAlertSupplier
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.view.ReadOnlyDict import ReadOnlyDict


class DIAObjectMissingError(Exception):
    """
    Raised when there is no DIAObject in the alert
    """

    ...


# Translate 2022-era field names to schema 7.x
_field_upgrades: dict[str, str] = {
    "midPointTai": "midpointMjdTai",
    "psFlux": "psfFlux",
    "psFluxErr": "psfFluxErr",
    "filterName": "band",
    "decl": "dec",
}


class LSSTAlertSupplier(BaseAlertSupplier):
    """
    Iterable class that, for each alert payload provided by the underlying alert_loader,
    returns an AmpelAlert instance.
    """

    # Override default
    deserialize: None | Literal["avro", "json"] = "avro"

    @staticmethod
    def _shape_dp(d: dict) -> ReadOnlyDict:
        return ReadOnlyDict(
            {_field_upgrades.get(k, k): v for k, v in d.items()}
        )

    @classmethod
    def _shape(cls, d: dict) -> AmpelAlertProtocol:
        if d["diaObject"]:
            diaObjectId = d["diaObject"]["diaObjectId"]
            dps = tuple(
                cls._shape_dp(dp)
                for dp in chain(
                    (d["diaSource"],),
                    d.get("prvDiaSources") or (),
                    d.get("prvDiaForcedSources") or (),
                    d.get("diaNondetectionLimit") or (),
                    ((d.get("diaObject"),) if d.get("diaObject") else ()),
                )
            )
            return AmpelAlert(
                id=d["alertId"],  # alert id
                stock=diaObjectId,  # internal ampel id
                datapoints=tuple(dps),
                extra={"kafka": kafka}
                if (kafka := d.get("__kafka"))
                else None,
            )
        raise DIAObjectMissingError

    def acknowledge(self, alerts: Iterator[AmpelAlertProtocol]) -> None:
        # invert transformation applied in _shape()
        self.alert_loader.acknowledge(
            {"__kafka": alert.extra["kafka"]}  # type: ignore[misc]
            for alert in alerts
            if alert.extra and "kafka" in alert.extra
        )

    def __next__(self) -> AmpelAlertProtocol:
        """
        :returns: a dict with a structure that AlertConsumer understands
        :raises StopIteration: when alert_loader dries out.
        :raises AttributeError: if alert_loader was not set properly before this method is called
        """
        d = self._deserialize(next(self.alert_loader))

        return self._shape(d)



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/view/LSSTT2Tabulator.py
# License           : BSD-3-Clause
# Author            : Marcus Fenner <mf@physik.hu-berlin.de>
# Date              : 25.05.2021
# Last Modified Date: 05.05.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Iterable, Sequence
from typing import Any

from astropy.table import Table

from ampel.abstract.AbsT2Tabulator import AbsT2Tabulator
from ampel.content.DataPoint import DataPoint
from ampel.types import StockId
from ampel.util.collections import ampel_iter

LSST_BANDPASSES = {
    "u": "lsstu",
    "g": "lsstg",
    "r": "lsstr",
    "i": "lssti",
    "z": "lsstz",
    "y": "lssty",
}


class LSSTT2Tabulator(AbsT2Tabulator):
    convert2jd: bool = True
    zp: float
    """ """

    def get_flux_table(
        self,
        dps: Iterable[DataPoint],
    ) -> Table:
        flux, fluxerr, filtername, tai = self.get_values(
            dps, ["psfFlux", "psfFluxErr", "band", "midpointMjdTai"]
        )
        if self.convert2jd:
            tai = self._to_jd(tai)
        filters = list(map(LSST_BANDPASSES.get, filtername))

        return Table(
            {
                "time": tai,
                "flux": flux,
                "fluxerr": fluxerr,
                "band": filters,
                # ZP for ELAsTiCC, might need to be corrected for LSST!
                "zp": [self.zp] * len(filters),
                "zpsys": ["ab"] * len(filters),
            },
            dtype=("float64", "float64", "float64", "str", "float64", "str"),
        )

    def get_positions(
        self, dps: Iterable[DataPoint]
    ) -> Sequence[tuple[float, float, float]]:
        return tuple(
            zip(
                self.get_jd(dps),
                *self.get_values(dps, ["ra", "dec"]),
                strict=False,
            )
        )

    def get_jd(self, dps: Iterable[DataPoint]) -> Sequence[float]:
        return self._to_jd(self.get_values(dps, ["midpointMjdTai"])[0])

    @staticmethod
    def _to_jd(dates: Sequence[Any]) -> Sequence[Any]:
        return [date + 2400000.5 for date in dates]

    def get_stock_id(self, dps: Iterable[DataPoint]) -> set[StockId]:
        return set(
            stockid
            for el in dps
            if "LSST" in el["tag"]
            for stockid in ampel_iter(el["stock"])
        )

    def get_stock_name(self, dps: Iterable[DataPoint]) -> list[str]:
        return [str(stock) for stock in self.get_stock_id(dps)]

    @staticmethod
    def get_values(
        dps: Iterable[DataPoint], params: Sequence[str]
    ) -> tuple[Sequence[Any], ...]:
        if tup := tuple(
            map(
                list,
                zip(
                    *(
                        [el["body"][param] for param in params]
                        for el in dps
                        if ("LSST_DP" in el["tag"] or "LSST_FP" in el["tag"])
                    ),
                    strict=False,
                ),
            )
        ):
            return tup
        return tuple([[]] * len(params))



#!/usr/bin/env python

import itertools
import uuid
from collections.abc import Iterable, Iterator
from typing import Annotated, Any, Literal

import confluent_kafka
from annotated_types import Gt, MinLen
from confluent_kafka.deserializing_consumer import DeserializingConsumer
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer

from ampel.abstract.AbsAlertLoader import AbsAlertLoader
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.log.AmpelLogger import AmpelLogger
from ampel.secret.NamedSecret import NamedSecret

from .HttpSchemaRepository import DEFAULT_SCHEMA
from .PlainAvroDeserializer import Deserializer, PlainAvroDeserializer


class SchemaRegistryURL(AmpelBaseModel):
    registry: str


class StaticSchemaURL(AmpelBaseModel):
    root_url: str = DEFAULT_SCHEMA


class SASLAuthentication(AmpelBaseModel):
    protocol: Literal["SASL_PLAINTEXT", "SASL_SSL"] = "SASL_PLAINTEXT"
    mechanism: Literal["PLAIN", "SCRAM-SHA-256", "SCRAM-SHA-512"] = (
        "SCRAM-SHA-512"
    )
    username: NamedSecret[str]
    password: NamedSecret[str]

    def librdkafka_config(self) -> dict[str, Any]:
        return {
            "security.protocol": self.protocol,
            "sasl.mechanism": self.mechanism,
            "sasl.username": self.username.get(),
            "sasl.password": self.password.get(),
        }


class KafkaAlertLoader(AbsAlertLoader[dict]):
    """
    Load alerts from one or more Kafka topics
    """

    #: Address of Kafka broker
    bootstrap: str
    #: Optional authentication
    auth: None | SASLAuthentication = None
    #: Topics to subscribe to
    topics: Annotated[list[str], MinLen(1)]
    #: Message schema (or url pointing to one)
    avro_schema: SchemaRegistryURL | StaticSchemaURL
    #: Consumer group name
    group_name: None | str = None
    #: time to wait for messages before giving up, in seconds
    timeout: Annotated[int, Gt(0)] = 1
    #: extra configuration to pass to confluent_kafka.Consumer
    kafka_consumer_properties: dict[str, Any] = {}

    def __init__(self, **kwargs):
        if isinstance(kwargs.get("avro_schema"), str):
            kwargs["avro_schema"] = {"root_url": kwargs["avro_schema"]}
        super().__init__(**kwargs)

        if isinstance(self.avro_schema, StaticSchemaURL):
            deserializer: Deserializer = PlainAvroDeserializer(
                self.avro_schema.root_url
            )
        else:
            deserializer = AvroDeserializer(
                SchemaRegistryClient({"url": self.avro_schema.registry})
            )

        config = (
            {
                "bootstrap.servers": self.bootstrap,
                "auto.offset.reset": "smallest",
                "enable.auto.commit": True,
                "enable.auto.offset.store": False,
                "auto.commit.interval.ms": 10000,
                "receive.message.max.bytes": 2**29,
                "enable.partition.eof": False,  # don't emit messages on EOF
                "value.deserializer": deserializer,
                "error_cb": self._raise_errors,
            }
            | (
                {
                    "group.id": self.group_name
                    if self.group_name
                    else str(uuid.uuid1())
                }
                if self.auth is None
                else self.auth.librdkafka_config()
                | {
                    "group.id": self.group_name
                    if self.group_name
                    else f"{self.auth.username.get()}-{uuid.uuid1()}",
                }
            )
            | self.kafka_consumer_properties
        )

        self._consumer = DeserializingConsumer(config)
        self._consumer.subscribe(self.topics)
        self._it = None

        self._poll_interval = max((1, min((3, self.timeout))))
        self._poll_attempts = max((1, int(self.timeout / self._poll_interval)))

    def _raise_errors(self, exc: Exception) -> None:
        raise exc

    def set_logger(self, logger: AmpelLogger) -> None:
        super().set_logger(logger)

    @staticmethod
    def _add_message_metadata(alert: dict, message: confluent_kafka.Message):
        meta = {}
        timestamp_kind, timestamp = message.timestamp()
        meta["timestamp"] = {
            (
                "create"
                if timestamp_kind == confluent_kafka.TIMESTAMP_CREATE_TIME
                else "append"
                if timestamp_kind == confluent_kafka.TIMESTAMP_LOG_APPEND_TIME
                else "unavailable"
            ): timestamp
        }
        meta["topic"] = message.topic()
        meta["partition"] = message.partition()
        meta["offset"] = message.offset()
        meta["key"] = message.key()

        alert["__kafka"] = meta
        return alert

    def acknowledge(self, alert_dicts: Iterable[dict]) -> None:
        """
        Store offsets of fully-processed messages
        """
        offsets: dict[tuple[str, int], int] = dict()
        for alert in alert_dicts:
            meta = alert["__kafka"]
            key, value = (meta["topic"], meta["partition"]), meta["offset"]
            if key not in offsets or value > offsets[key]:
                offsets[key] = value
        try:
            self._consumer.store_offsets(
                offsets=[
                    confluent_kafka.TopicPartition(
                        topic, partition, offset + 1
                    )
                    for (topic, partition), offset in offsets.items()
                ]
            )
        except confluent_kafka.KafkaException as exc:
            # librdkafka will refuse to store offsets on a partition that is not
            # currently assigned. this can happen if the group is rebalanced
            # while a batch of messages is in flight. see also:
            # https://github.com/confluentinc/confluent-kafka-dotnet/issues/1861
            err = exc.args[0]
            if err.code() != confluent_kafka.KafkaError._STATE:  # noqa: SLF001
                raise

    def _poll(self) -> confluent_kafka.Message | None:
        """
        Poll for a message, ignoring nonfatal errors
        """
        message = None
        for _ in range(self._poll_attempts):
            # wake up occasionally to catch SIGINT
            message = self._consumer.poll(self._poll_interval)
            if message is not None:
                if err := message.error():
                    if (
                        err.code()
                        == confluent_kafka.KafkaError.UNKNOWN_TOPIC_OR_PART
                    ):
                        # ignore unknown topic messages
                        continue
                    if err.code() in (
                        confluent_kafka.KafkaError._TIMED_OUT,  # noqa: SLF001
                        confluent_kafka.KafkaError._MAX_POLL_EXCEEDED,  # noqa: SLF001
                    ):
                        # bail on timeouts
                        if self._logger:
                            self._logger.debug(f"Got {err}")
                        return None
                break
        if message is None:
            return message
        if message.error():
            raise message.error()
        return message

    def _consume(self) -> Iterator[dict]:
        while True:
            message = self._poll()
            if message is None:
                return
            else:
                yield self._add_message_metadata(message.value(), message)

    def alerts(self, limit: None | int = None) -> Iterator[dict]:
        """
        Generate alerts until timeout is reached
        :returns: dict instance of the alert content
        :raises StopIteration: when no alerts recieved within timeout
        """

        yield from itertools.islice(self._consume(), limit)

    def __next__(self) -> dict:
        if self._it is None:
            self._it = self.alerts()
        return next(self._it)



from ampel.model.UnitModel import UnitModel

from .LSSTAlertConsumerTemplate import LSSTAlertConsumerTemplate


class ElasticcAlertConsumerTemplate(LSSTAlertConsumerTemplate):
    loader: str | UnitModel = UnitModel(
        unit="KafkaAlertLoader",
        config={
            "bootstrap": "public.alerts.ztf.uw.edu:9092",
            "group_name": "ampel-test",
            "topics": ["elasticc-test-only-3"],
            "timeout": 30,
        },
    )
